@@ -2,12 +2,14 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.i18n import gettext
-from trytond.model import ModelSQL, ModelView, Workflow, fields, Unique
+from trytond.model import Index, ModelSQL, ModelView, Unique, Workflow, fields
 from trytond.model.exceptions import AccessError
-from trytond.pool import PoolMeta, Pool
+from trytond.modules.account.exceptions import AccountMissing
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool, PoolMeta
 from trytond.pyson import Eval, If
-from trytond.transaction import Transaction
-from trytond.wizard import Wizard, StateTransition
+from trytond.transaction import Transaction, check_access
+from trytond.wizard import StateTransition, Wizard
 
 
 class Configuration(metaclass=PoolMeta):
@@ -42,15 +44,13 @@ class ConfigurationDefaultAccount(metaclass=PoolMeta):
         domain=[
             ('type.statement', '=', 'balance'),
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
+            ])
     deferred_account_expense = fields.Many2One(
         'account.account', "Deferred Account Expense",
         domain=[
             ('type.statement', '=', 'balance'),
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
+            ])
 
 
 class InvoiceDeferred(Workflow, ModelSQL, ModelView):
@@ -60,26 +60,27 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
     _states = {
         'readonly': Eval('state') != 'draft',
         }
-    _depends = ['state']
 
     company = fields.Many2One(
         'company.company', "Company", required=True,
         states={
             'readonly': (Eval('state') != 'draft') & Eval('invoice_line'),
-            },
-        depends=['state', 'invoice_line'])
+            })
     type = fields.Selection([
             ('out', "Customer"),
             ('in', "Supplier"),
             ], "Type", required=True,
-        states=_states, depends=_depends)
+        states=_states)
     journal = fields.Many2One(
-        'account.journal', "Journal", required=True,
-        states=_states,
+        'account.journal', "Journal",
+        states={
+            'readonly': _states['readonly'],
+            'required': Eval('state') != 'draft',
+            },
         context={
             'company': Eval('company', -1),
             },
-        depends=_depends + ['company'])
+        depends={'company'})
     invoice_line = fields.Many2One(
         'account.invoice.line', "Invoice Line", required=True,
         domain=[
@@ -88,22 +89,22 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
             ('invoice.state', 'in', ['posted', 'paid']),
             ('invoice.company', '=', Eval('company', -1)),
             ],
-        states=_states, depends=_depends + ['type', 'company'])
-    amount = fields.Numeric(
-        "Amount", digits=(16, Eval('currency_digits', 2)), required=True,
-        states=_states, depends=_depends + ['currency_digits'])
+        states=_states)
+    amount = Monetary(
+        "Amount", currency='currency', digits='currency', required=True,
+        states=_states)
     start_date = fields.Date(
         "Start Date", required=True,
         domain=[
             ('start_date', '<', Eval('end_date', None)),
             ],
-        states=_states, depends=_depends + ['end_date'])
+        states=_states)
     end_date = fields.Date(
         "End Date", required=True,
         domain=[
             ('end_date', '>', Eval('start_date', None)),
             ],
-        states=_states, depends=_depends + ['start_date'])
+        states=_states)
     moves = fields.One2Many(
         'account.move', 'origin', "Moves", readonly=True,
         order=[
@@ -113,14 +114,12 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
             ('draft', "Draft"),
             ('running', "Running"),
             ('closed', "Closed"),
-            ], "State", readonly=True, required=True)
+            ], "State", readonly=True, required=True, sort=False)
 
     currency = fields.Function(fields.Many2One(
             'currency.currency', "Currency"), 'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer("Currency Digits"),
-        'on_change_with_currency_digits')
 
-    del _states, _depends
+    del _states
 
     @classmethod
     def __setup__(cls):
@@ -130,12 +129,15 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
             ('invoice_line_unique', Unique(table, table.invoice_line),
                 'account_invoice_defer.msg_defer_invoice_line_unique'),
             ]
+        cls._sql_indexes.add(
+            Index(
+                table, (table.state, Index.Equality()),
+                where=table.state.in_(['draft', 'running'])))
         cls.journal.domain = [
             If(Eval('type') == 'out',
                 ('type', 'in', cls._journal_types('out')),
                 ('type', 'in', cls._journal_types('in'))),
             ]
-        cls.journal.depends += ['type']
         cls._transitions |= set((
                 ('draft', 'running'),
                 ('running', 'closed'),
@@ -147,6 +149,14 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
                     },
                 })
 
+    @classmethod
+    def __register__(cls, module_name):
+        table_h = cls.__table_handler__(module_name)
+        super().__register__(module_name)
+
+        # Migration from 6.6: drop not null on journal
+        table_h.not_null_action('journal', 'remove')
+
     @classmethod
     def _journal_types(cls, type):
         if type == 'out':
@@ -154,15 +164,20 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
         else:
             return ['expense']
 
-    @fields.depends('type')
+    @fields.depends(methods=['set_journal'])
     def on_change_type(self):
-        Journal = self.__class__.journal.get_target()
-        journal_types = self.__class__._journal_types(self.type)
-        journals = Journal.search([
-                ('type', 'in', journal_types),
-                ], limit=1)
-        if journals:
-            self.journal, = journals
+        self.set_journal()
+
+    @fields.depends('type')
+    def set_journal(self, pattern=None):
+        pool = Pool()
+        Journal = pool.get('account.journal')
+        pattern = pattern.copy() if pattern is not None else {}
+        pattern.setdefault('type', {
+                'out': 'revenue',
+                'in': 'expense',
+                }.get(self.type))
+        self.journal = Journal.find(pattern)
 
     @fields.depends('invoice_line', 'start_date', 'company')
     def on_change_invoice_line(self):
@@ -180,19 +195,17 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
             else:
                 self.amount = self.invoice_line.amount
 
+    @classmethod
+    def default_company(cls):
+        return Transaction().context.get('company')
+
     @classmethod
     def default_state(cls):
         return 'draft'
 
     @fields.depends('company')
     def on_change_with_currency(self, name=None):
-        if self.company:
-            return self.company.currency.id
-
-    @fields.depends('company')
-    def on_change_with_currency_digits(self, name=None):
-        if self.company:
-            return self.company.currency.digits
+        return self.company.currency if self.company else None
 
     @classmethod
     @ModelView.button
@@ -202,7 +215,7 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
         Period = pool.get('account.period')
         # Ensure it starts at an opened period
         for deferral in deferrals:
-            Period.find(deferral.company.id, deferral.start_date)
+            Period.find(deferral.company, deferral.start_date)
         # Set state before create moves and defer amount to pass assert
         cls.write(deferrals, {'state': 'running'})
         cls.create_moves(deferrals)
@@ -354,9 +367,17 @@ class InvoiceDeferred(Workflow, ModelSQL, ModelView):
         if invoice.type == 'out':
             balance.account = configuration.get_multivalue(
                 'deferred_account_revenue', company=self.company.id)
+            if not balance.account:
+                raise AccountMissing(gettext(
+                        'account_invoice_defer.'
+                        'msg_missing_deferred_account_revenue'))
         else:
             balance.account = configuration.get_multivalue(
                 'deferred_account_expense', company=self.company.id)
+            if not balance.account:
+                raise AccountMissing(gettext(
+                        'account_invoice_defer.'
+                        'msg_missing_deferred_account_expense'))
         balance.debit, balance.credit = income.credit, income.debit
         if balance.account.party_required:
             balance.party = invoice.party
@@ -374,7 +395,7 @@ class InvoiceDeferredCreateMoves(Wizard):
     def transition_create_moves(self):
         pool = Pool()
         InvoiceDeferred = pool.get('account.invoice.deferred')
-        with Transaction().set_context(_check_access=True):
+        with check_access():
             deferrals = InvoiceDeferred.search([
                     ('state', '=', 'running'),
                     ])
@@ -479,8 +500,7 @@ class InvoiceLine(metaclass=PoolMeta):
         states={
             'readonly': Eval('invoice_state') != 'draft',
             'invisible': ~Eval('deferrable', False),
-            },
-        depends=['deferrable', 'defer_to', 'invoice_state'])
+            })
     defer_to = fields.Date(
         "Defer To",
         domain=[
@@ -491,8 +511,7 @@ class InvoiceLine(metaclass=PoolMeta):
         states={
             'readonly': Eval('invoice_state') != 'draft',
             'invisible': ~Eval('deferrable', False),
-            },
-        depends=['deferrable', 'defer_from', 'invoice_state'])
+            })
 
     @fields.depends('product')
     def on_change_with_deferrable(self, name=None):
